
import numpy as np


from replay_buffers import AbstractBuffer


class SimpleBuffer(AbstractBuffer):
    """
    """
    def __init__(self, settings):
        """
        """        
        self.buffer_size = settings.buffer_settings["buffer_size"]
        self.buffer_list = []
    
    def add(self, experience):
        """
        """
        len_buffer = len(self.buffer_list)
        len_exp = len(experience)
        #
        if len_buffer + len_exp >= self.buffer_size:
            self.buffer[0:len_buffer + len_exp - self.buffer_size] = []
        #
        self.buffer_list.extend(experience)
        #
            
    def sample(self, size):
        """
        """
        return np.random.sample(self.buffer_list, size)







from agents import AbstractAgent

class DQN(AbstractAgent):
    """
    """
    def __init__(self, settings):
        """
        """
        self.settings = settings

    def build_q_net(self):
        """
        """
        pass

    #
    def generate(self, env):
        """ generate experience
        """
        pass

    #
    def get_batch(self, buffer):
        """ sample a batch from buffer
        """
        pass

    def optimize(self, batch):
        """ optimization step
        """
        pass



    




from utils import get_package_root_dir

dir_pkg_root = get_package_root_dir()
print(dir_pkg_root)


from utils.basic_settings import BasicSettings
settings = BasicSettings()
settings.env_settings = {}
settings.env_settings["size"] = 5
settings.env_settings["partial"] = False

from envs.grid_world import GridWorld
env = GridWorld(settings)
env.display()


    




from abc import ABCMeta, abstractmethod

class AbstractBuffer(metaclass=ABCMeta):
    """
    """
    def __init__(self, settings):
        """
        """
        pass

    @abstractmethod    
    def add(self, experience):
        """
        """
        pass

    @abstractmethod             
    def sample(self, num):
        """
        """
        pass





import numpy as np
import itertools
import scipy.misc
import matplotlib.pyplot as plt

from envs import AbstractEnv


class GridItem():
    def __init__(self,coordinates,size,intensity,channel,reward,name):
        self.x = coordinates[0]
        self.y = coordinates[1]
        self.size = size
        self.intensity = intensity
        self.channel = channel
        self.reward = reward
        self.name = name
        
class GridWorld(AbstractEnv):
    """
    """
    def __init__(self, settings):
        """
        """
        self.settings = settings
        self.sizeX = self.settings.env_settings["size"]
        self.sizeY = self.settings.env_settings["size"]
        self.actions = 4
        self.objects = []
        self.partial = self.settings.env_settings["partial"]
        #
        self.reset()        
        #

    def display(self):
        """
        """
        plt.figure()
        plt.imshow(self.map_to_pic(self.state), interpolation="nearest")
        plt.show()
        #
        
    def reset(self):
        """
        """
        self.objects = []
        #
        hero = GridItem(self.get_new_posi(),1,1,2,None,'hero')
        self.objects.append(hero)
        #
        food = GridItem(self.get_new_posi(),1,1,1,1,'food')
        self.objects.append(food)
        food2 = GridItem(self.get_new_posi(),1,1,1,1,'food')
        self.objects.append(food2)
        food3 = GridItem(self.get_new_posi(),1,1,1,1,'food')
        self.objects.append(food3)
        food4 = GridItem(self.get_new_posi(),1,1,1,1,'food')
        self.objects.append(food4)
        #
        pois = GridItem(self.get_new_posi(),1,1,0,-1,'fire')
        self.objects.append(pois)
        pois2 = GridItem(self.get_new_posi(),1,1,0,-1,'fire')
        self.objects.append(pois2)        
        #    
        return self.render_world()
        #
    
    def render_world(self):
        """
        """
        a = np.ones([self.sizeY+2, self.sizeX+2, 3])
        a[1:-1,1:-1,:] = 0
        hero = None
        for item in self.objects:
            a[item.y+1:item.y+item.size+1,\
              item.x+1:item.x+item.size+1,\
              item.channel] = item.intensity
            if item.name == 'hero':
                hero = item
        if self.partial == True:
            a = a[hero.y:hero.y+3,hero.x:hero.x+3,:]                    
        #
        self.state = a
        return a
        #
    
    def map_to_pic(self, state):
        """
        """
        b = scipy.misc.imresize(state[:,:,0],[84,84,1],interp='nearest')
        c = scipy.misc.imresize(state[:,:,1],[84,84,1],interp='nearest')
        d = scipy.misc.imresize(state[:,:,2],[84,84,1],interp='nearest')
        pic = np.stack([b,c,d], axis=2)
        #        
        return pic
    
    def get_new_posi(self):
        """
        """
        iterables = [range(self.sizeX), range(self.sizeY)]
        points = []
        for t in itertools.product(*iterables):
            points.append(t)
        current_positions = []
        for object_a in self.objects:
            if (object_a.x,object_a.y) not in current_positions:
                current_positions.append((object_a.x,object_a.y))
        for pos in current_positions:
            points.remove(pos)
        location = np.random.choice(range(len(points)),replace=False)
        return points[location]
        
    #
    def execute_action(self, action):
        """
        """
        penalty = self.move_char(action)
        reward, done = self.check_reward()
        state = self.render_world()
        #
        return state, (reward+penalty), done
        
    def move_char(self, direction):
        """
        """
        # 0 - up, 1 - down, 2 - left, 3 - right
        hero = self.objects[0]
        heroX = hero.x
        heroY = hero.y
        penalize = 0.0
        if direction == 0 and hero.y >= 1:
            hero.y -= 1
        if direction == 1 and hero.y <= self.sizeY-2:
            hero.y += 1
        if direction == 2 and hero.x >= 1:
            hero.x -= 1
        if direction == 3 and hero.x <= self.sizeX-2:
            hero.x += 1     
        if hero.x == heroX and hero.y == heroY:
            penalize = 0.0
        self.objects[0] = hero
        return penalize
    
    def check_reward(self):
        """
        """
        others = []
        for obj in self.objects:
            if obj.name == 'hero':
                hero = obj
            else:
                others.append(obj)
        #
        for other in others:
            if hero.x == other.x and hero.y == other.y:
                self.objects.remove(other)
                if other.reward == 1:
                    self.objects.append(GameItem(self.get_new_posi(),1,1,1,1,'food'))
                else: 
                    self.objects.append(GameItem(self.get_new_posi(),1,1,0,-1,'fire'))
                return other.reward, False
        else:
            return 0.0, False
        #

#





class SimpleTrainer():
    """
    """
    def __init__(self, settings, agent, env, buffer):
        """
        """
        self.settings = settings
        self.agent = agent
        self.env = env
        self.buffer = buffer

    def train(self):
        """
        """
        num_boost = self.settings.trainer_settings["num_boost"]
        num_gen = self.settings.trainer_settings["num_gen"]
        num_optim = self.settings.trainer_settings["num_optim"]
        #
        for idx_boost in range(num_boost):
            # generate experience
            for idx_gen in range(num_gen):
                experience = self.agent.generate(self.env)
                self.buffer.add(experience)
            # optimize
            for idx_optim in range(num_optim):
                batch = self.agent.get_batch(self.buffer)
                self.agent.optimize(batch)
            #
        #




    





import os
from utils.basic_settings import BasicSettings


#
env_type = "grid_world"
agent_type = "dqn"
settings_filename = "settings_dqn.json"
#
buffer_type = "simple_buffer"
trainer_type = "simple_trainer"
#
##
dir_data_root = "./data_root"
dir_rel_settings = "settings"
#
print(os.path.abspath(dir_data_root))
#
settings_filepath = os.path.join(dir_data_root, dir_rel_settings, settings_filename)
#

##
if env_type == "grid_world":
    from envs.grid_world import GridWorld as Env
#
if agent_type == "dqn":
    from agents.dqn import DQN as Agent
#

##
if buffer_type == "simple_buffer":
    from replay_buffers.simple_buffer import SimpleBuffer as Buffer
#
if trainer_type == "simple_trainer":
    from trainers.simple_trainer import SimpleTrainer as Trainer
#

#
assert False, ""
#

##
settings = BasicSettings(settings_filepath)
settings.dir_base = dir_data_root
settings.dir_rel_settings = dir_rel_settings
#
env = Env(settings)
buffer = Buffer(settings)
agent = Agent(settings)
#
trainer = Trainer(settings, agent, env, buffer)
#
trainer.train()
#




import os
import time
import json
import logging

class BasicSettings():
    """
    """
    def __init__(self, file_path=None):
        """
        """
        self.dir_base = "./data_root"
        self.dir_rel_log = "log"
        self.dir_rel_settings = "settings"
        self.dir_rel_model = "stat_dict"
        self.log_path = None
        #
        self.agent_settings = {"name": "agentname"}
        self.env_settings = {"name": "envname"}
        self.buffer_settings = {"buffer_size": 5000}
        self.trainer_settings = {}
        #
        self.trainer_settings["num_boost"] = 1000
        self.trainer_settings["num_gen"] = 100
        self.trainer_settings["num_optim"] = 100
        #
        self.load_from_json_file(file_path)
        #
        # except from saving
        self.except_list = ["dir_base", "dir_log", "dir_settings", "dir_model",
               "log_path", "dir_rel_log", "dir_rel_settings", "dir_rel_model",
               "except_list"]
        #

    #
    def check_settings(self):
        """
        """
        # directories
        self.dir_log = os.path.join(self.dir_base, self.dir_rel_log)
        self.dir_settings = os.path.join(self.dir_base, self.dir_rel_settings)
        self.dir_model = os.path.join(self.dir_base, self.dir_rel_model)
        #        
        if not os.path.exists(self.dir_base): os.mkdir(self.dir_base)
        if not os.path.exists(self.dir_log): os.mkdir(self.dir_log)
        if not os.path.exists(self.dir_settings): os.mkdir(self.dir_settings)
        if not os.path.exists(self.dir_model): os.mkdir(self.dir_model)
        #
        # logger
        str_datetime = time.strftime("%Y-%m-%d-%H-%M-%S")
        if self.log_path is None: self.log_path = os.path.join(
                self.dir_log, "log_%s_%s_%s.txt" % (self.agent_settings["name"],
                self.env_settings["name"], str_datetime) )
        #
        self.logger = self.create_logger(self.log_path)
        print("settings checked")
        #
        self.logger.info(self.trans_info_to_dict())
        # self.display()
        #

    def create_logger(self, log_path):
        """
        """
        logger = logging.getLogger(log_path)  # use log_path as log_name
        logger.setLevel(logging.INFO)
        handler = logging.FileHandler(log_path, encoding='utf-8') 
        handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        # formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        # self.logger.info('test')
        return logger

    def create_or_reset_log_file(self, log_path=None):
        if log_path is None:
            log_path = self.log_path        
        with open(log_path, 'w', encoding='utf-8'):
            print("log file reset")
    
    def close_logger(self, logger=None):
        if logger is None:
            logger = self.logger
        for item in logger.handlers:
            item.close()
            print("logger handler item closed")

    #
    def display(self):
        """
        """        
        print()
        for name, value in vars(self).items():
            if not isinstance(value, (int, float, str, bool, list, dict, tuple)):
                continue
            print(str(name) + ': ' + str(value))
        print()
    
    #
    def trans_info_to_dict(self):
        """
        """                
        info_dict = {}
        for name, value in vars(self).items():
            if not isinstance(value, (int, float, str, bool, list, dict, tuple)):
                continue
            if str(name) in self.except_list:
                continue
            info_dict[str(name)] = value        
        return info_dict
    
    def assign_info_from_dict(self, info_dict):
        """
        """
        for key in info_dict:
            value = info_dict[key]
            setattr(self, key, value)
        #

    def assign_info_from_namedspace(self, named_data):
        """
        """
        for key in named_data.__dict__.keys():                 
            self.__dict__[key] = named_data.__dict__[key]
        #
        
    def save_to_json_file(self, file_path):
        """
        """
        info_dict = self.trans_info_to_dict()
        #
        with open(file_path, "w", encoding="utf-8") as fp:
            json.dump(info_dict, fp, ensure_ascii=False, indent=4)
        #       
    
    def load_from_json_file(self, file_path):
        """
        """
        if file_path is None:
            print("settings file: %s NOT found, using default settings" % file_path)
            return
        #
        with open(file_path, "r", encoding="utf-8") as fp:
            info_dict = json.load(fp)
        #
        self.assign_info_from_dict(info_dict)
        #


   
#       
if __name__ == '__main__':
    
    sett = BasicSettings()
    
    #print(dir(sett))
    #l = [i for i in dir(sett) if inspect.isbuiltin(getattr(sett, i))]
    #l = [i for i in dir(sett) if inspect.isfunction(getattr(sett, i))]
    #l = [i for i in dir(sett) if not callable(getattr(sett, i))]
    
    sett.check_settings()
    sett.display()
    
    print(sett.__dict__.keys())
    print()
    
    info_dict = sett.trans_info_to_dict()
    print(info_dict)
    print()
    
    #
    sett.assign_info_from_dict(info_dict)
    
    #
    info_dict = sett.trans_info_to_dict()
    print(info_dict)
    print()
    #
    
    file_path = os.path.join(sett.dir_settings, "settings_template.json")
    sett.save_to_json_file(file_path)    
    sett.load_from_json_file(file_path)
    
    #
    info_dict = sett.trans_info_to_dict()
    print(info_dict)
    print()
    #

    #    
    sett.close_logger()
    #

    

    

    




from abc import ABCMeta, abstractmethod

class AbstractAgent(metaclass=ABCMeta):
    """
    """
    def __init__(self, settings):
        """
        """
        pass

    @abstractmethod
    def generate(self, env):
        """ generate experience
        """
        pass

    @abstractmethod
    def get_batch(self, buffer):
        """ sample a batch from buffer
        """
        pass

    @abstractmethod
    def optimize(self, batch):
        """ optimization step
        """
        pass

    





from abc import ABCMeta, abstractmethod

class AbstractTrainer(metaclass=ABCMeta):
    """
    """
    def __init__(self, settings, agent, env, buffer):
        """
        """
        pass

    @abstractmethod
    def train(self):
        """
        """
        pass




from abc import ABCMeta, abstractmethod

class AbstractEnv(metaclass=ABCMeta):
    """
    """
    def __init__(self, settings):
        """
        """
        pass

    @abstractmethod    
    def execute_action(self, action):
        """
        """
        pass




# -*- coding: utf-8 -*-

from __future__ import with_statement

import sys
if sys.version_info < (3, 6):
    sys.exit('Python 3.6 or greater is required.')

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
    
import setuptools

import ziqiang


with open('README.md') as fp:
    readme = fp.read()

with open('LICENSE') as fp:
    license = fp.read()

setup(name = 'ziqiang',
      version = ziqiang.__version__,
      description = 'A package for reinforcement learning algorithms.',
      url='https://github.com/Li-Ming-Fan/ziqiang',
      #
      long_description = readme,
      long_description_content_type="text/markdown",
      author = 'Ming-Fan Li',
      author_email = 'li_m_f@163.com',
      maintainer='Ming-Fan Li',
      maintainer_email='li_m_f@163.com',      
      packages=setuptools.find_packages(),
      #license=license,
      platforms=['any'],
      classifiers=[]
      )




import os

def get_package_root_dir():
    """
    """
    file_path = os.path.abspath(__file__)
    file_dir = os.path.dirname(file_path)
    dir_root = file_dir.rstrip("utils")
    # print(file_path)
    # print(file_dir)
    # print(dir_root)
    return dir_root
    


